`timescale 1ns / 1ps

//******************************************************************************
//* MiniRISC CPU v2.0                                                          *
//*                                                                            *
//* A processzor adatstruktrjt vezrl egysg.                              *
//******************************************************************************
module control_unit(
   //rajel s reset.
   input  wire        clk,             //rajel
   input  wire        rst,             //Aszinkron reset
   
   //A programmemrival kapcsolatos jelek.
   output wire [7:0]  prg_mem_addr,    //Cmbusz
   input  wire [15:0] prg_mem_din,     //Olvassi adatbusz
   
   //Az adatmemrival kapcsolatos jelek.
   output wire        bus_req,         //Busz hozzfrs krse
   input  wire        bus_grant,       //Busz hozzfrs megadsa
   output wire        data_mem_wr,     //rs engedlyez jel
   output wire        data_mem_rd,     //Olvass engedlyez jel
   
   //Az utastsban lv konstans adat.
   output wire [7:0]  const_data,
   
   //Az adatstruktra multiplexereinek vezrl jelei.
   output wire        wr_data_sel,     //A regiszterbe rand adat kivlasztsa
   output wire        addr_op2_sel,    //Az ALU 2. operandusnak kivlasztsa
   
   //A regisztertmbbel kapcsolatos jelek.
   output wire        reg_wr_en,       //rs engedlyez jel
   output wire [3:0]  reg_addr_x,      //Regiszter cme (X port)
   output wire [3:0]  reg_addr_y,      //Regiszter cme (Y port)
   
   //Az ALU-val kapcsolatos jelek.
   output wire [1:0]  alu_op_type,     //ALU mvelet kivlaszt jel
   output wire [1:0]  alu_arith_sel,   //Aritmetikai mvelet kivlaszt jel
   output wire [1:0]  alu_logic_sel,   //Logikai mvelet kivlaszt jel
   output wire [3:0]  alu_shift_sel,   //Shiftelsi mvelet kivlaszt jel
   output wire [3:0]  alu_flag_din,    //A flag-ekbe rand rtk
   output wire        alu_flag_wr,     //A flag-ek rs engedlyez jele
   input  wire        alu_flag_z,      //Zero flag
   input  wire        alu_flag_c,      //Carry flag
   input  wire        alu_flag_n,      //Negative flag
   input  wire        alu_flag_v,      //Overflow flag
   
   //Ugrsi cm az adatstruktrtl.
   input  wire [7:0]  jump_addr,
   
   //Megszaktskr bemenet (aktv magas szintrzkeny).
   input  wire        irq,
   
   //A debug interfsz jelei.
   input  wire [7:0]  dbg_data_in,     //Adatbemenet
   input  wire        dbg_break,       //Program vgrehajtsnak felfggesztse
   input  wire        dbg_continue,    //Program vgrehajtsnak folytatsa
   input  wire        dbg_pc_wr,       //A programszmll rs engedlyez jele
   input  wire        dbg_flag_wr,     //A flag-ek rs engedlyez jele
   input  wire        dbg_reg_wr,      //A regisztertmb rs engedlyez jele
   input  wire        dbg_mem_wr,      //Az adatmemria rs engedlyez jele
   input  wire        dbg_mem_rd,      //Az adatmemria olvass engedlyez jele
   output wire        dbg_instr_dec,   //Az utasts dekdols jelzse
   output wire        dbg_int_req,     //A megszakts kiszolglsnak jelzse
   output wire        dbg_is_brk,      //A trspont llapot jelzse
   output wire        dbg_flag_ie,     //Megyszakts engedlyez flag (IE)
   output wire        dbg_flag_if,     //Megyszakts flag (IF)
   output wire [13:0] dbg_stack_top    //A verem tetejn lv adat
);

`include "src\MiniRISC_CPU\control_defs.vh"
`include "src\MiniRISC_CPU\opcode_defs.vh"

//******************************************************************************
//* Vezrl jelek.                                                             *
//******************************************************************************
wire initialize;                       //Inicializls
wire fetch;                            //Utasts lehvs
wire ex_jump;                          //Ugrs vgrehajtsa
wire ex_call;                          //Szubrutinhvs vgrehajtsa
wire ex_ret_sub;                       //Visszatrs szubrutinbl
wire ex_ret_int;                       //Visszatrs megszaktsbl


//******************************************************************************
//* Programszmll (PC). A lehvand utasts cmt trolja.                  *
//******************************************************************************
reg  [7:0] pc;                         //Programszmll regiszter
wire [7:0] return_addr;                //Visszatrsi cm

always @(posedge clk)
begin
   if (initialize)
      pc <= RST_VECTOR;                //A reset vektor betltse
   else
      if (dbg_int_req)
         pc <= INT_VECTOR;             //A megszakts vektor betltse
      else
         if (dbg_is_brk && dbg_pc_wr)
            pc <= dbg_data_in;         //A debug modul rja a programszmllt
         else
            if (ex_jump || ex_call)
               pc <= jump_addr;        //Az ugrsi cm betltse
            else
               if (ex_ret_sub || ex_ret_int)
                  pc <= return_addr;   //A visszatrsi cm betltse
               else
                  if (fetch)
                     pc <= pc + 8'd1;  //A programszmll nvelse
end

//A programmemria cmbusznak meghajtsa.
assign prg_mem_addr = pc;


//******************************************************************************
//* Verem. Szubrutinhvs s megszaktskrs esetn ide mentdik el a         *
//* programszmll, valamint a flag-ek rtke.                                *
//******************************************************************************
wire [13:0] stack_din;                 //A verembe rand adat
wire [13:0] stack_dout;                //A verem tetejn lv adat

stack #(
   //Az adat szlessge bitekben.
   .DATA_WIDTH(14)
) stack (
   //rajel.
   .clk(clk),
   
   //Adatvonalak.
   .data_in(stack_din),                //A verembe rand adat
   .data_out(stack_dout),              //A verem tetejn lv adat
   
   //Vezrl bemenetek.
   .push(dbg_int_req | ex_call),       //Adat rsa a verembe
   .pop(ex_ret_sub | ex_ret_int)       //Adat olvassa a verembl
);

//A verembe elmentjk a programszmllt s az ALU flag-eket.
assign stack_din[7:0] = pc;
assign stack_din[8]   = alu_flag_z;
assign stack_din[9]   = alu_flag_c;
assign stack_din[10]  = alu_flag_n;
assign stack_din[11]  = alu_flag_v;
assign stack_din[12]  = dbg_flag_ie;
assign stack_din[13]  = dbg_flag_if;

//A visszatrsi cm.
assign return_addr    = stack_dout[7:0];

//Az ALU flag-ekkel kapcsolatos jelek. Break llapotban a debug
//modul rhatja a flag-eket, egybknt pedig a verembe elmentett
//rtkek llthatk vissza.
assign alu_flag_wr    = (dbg_is_brk) ? dbg_flag_wr      : ex_ret_int;
assign alu_flag_din   = (dbg_is_brk) ? dbg_data_in[3:0] : stack_dout[11:8];

//A verem tetejn lv adat.
assign dbg_stack_top  = stack_dout;


//******************************************************************************
//* Utastsregiszter. A programmemribl lehvott utastst trolja.         *
//* Az utastsok formtumt lsd az 'opcode_defs.vh' fjlban.                 *
//******************************************************************************
reg [15:0] ir;

always @(posedge clk)
begin
   if (fetch)
      ir <= prg_mem_din;
end

//Az ALU msodik operandusnak kivlasz jele:
//0: konstans  / abszolt cmzs
//1: regiszter / indirekt cmzs
assign addr_op2_sel  = (ir[15:12] == REG_OP_PREFIX);

//Az utastsban trolt mveleti kd.
wire [3:0] opcode    = (addr_op2_sel) ? ir[7:4] : ir[15:12];

//A regiszterek cmei.
assign reg_addr_x    = ir[11:8];
assign reg_addr_y    = ir[3:0];

//Az utastsban lv konstans adat.
assign const_data    = ir[7:0];

//Az ALU mveletek kivlaszt jelei.
assign alu_arith_sel = opcode[1:0];
assign alu_logic_sel = opcode[1:0];
assign alu_shift_sel = ir[3:0];

//A programvezrlsi mvelet kdja.
wire [3:0] ctrl_op   = ir[11:8];


//******************************************************************************
//* A processzort vezrl llapotgp.                                          *
//******************************************************************************
controller_fsm controller_fsm(
   //rajel s reset.
   .clk(clk),                          //rajel
   .rst(rst),                          //Aszinkron reset
   
   //A lehvott utastssal kapcsolatos jelek.
   .addr_op2_sel(addr_op2_sel),        //Az ALU 2. operandusnak kivlasztsa
   .opcode(opcode),                    //Az utastsban lv mveleti kd
   .ctrl_op(ctrl_op),                  //A programvezrlsi mvelet kdja
   
   //A processzor llapotval kapcsolatos jelek.
   .initialize(initialize),            //Inicializls
   .fetch(fetch),                      //Utasts lehvs
   .decode(dbg_instr_dec),             //Utasts dekdols
   .interrupt(dbg_int_req),            //Megszakts kiszolgls
   
   //A programvezrl utastsok vgrehajtsval kapcsolatos jelek.
   .ex_jump(ex_jump),                  //Ugrs vgrehajtsa
   .ex_call(ex_call),                  //Szubrutinhvs vgrehajtsa
   .ex_ret_sub(ex_ret_sub),            //Visszatrs szubrutinbl
   .ex_ret_int(ex_ret_int),            //Visszatrs megszaktsbl
   
   //Az adatstruktrval kapcsolatos jelek.
   .wr_data_sel(wr_data_sel),          //A regiszterbe rand adat kivlasztsa
   .reg_wr_en(reg_wr_en),              //A regisztertmb rs engedlyez jele
   .alu_op_type(alu_op_type),          //ALU mvelet kivlaszt jel
   .alu_flag_z(alu_flag_z),            //Zero flag
   .alu_flag_c(alu_flag_c),            //Carry flag
   .alu_flag_n(alu_flag_n),            //Negative flag
   .alu_flag_v(alu_flag_v),            //Overflow flag
   
   //Az adatmemrival kapcsolatos jelek.
   .bus_req(bus_req),                  //Busz hozzfrs krse
   .bus_grant(bus_grant),              //Busz hozzfrs megadsa
   .data_mem_wr(data_mem_wr),          //rs engedlyez jel
   .data_mem_rd(data_mem_rd),          //Olvass engedlyez jel
   
   //A megszaktssal kapcsolatos jelek.
   .irq(irq),                          //Megszaktskr bemenet
   .flag_ie_din(stack_dout[12]),       //Az IE flag-ba rand rtk
   .flag_ie(dbg_flag_ie),              //Megyszakts engedlyez flag (IE)
   .flag_if_din(stack_dout[13]),       //Az IE flag-ba rand rtk
   .flag_if(dbg_flag_if),              //Megyszakts flag (IF)
   
   //A debug interfsz jelei.
   .dbg_break(dbg_break),              //Program vgrehajtsnak felfggesztse
   .dbg_continue(dbg_continue),        //Program vgrehajtsnak folytatsa
   .dbg_ie_wr(dbg_flag_wr),            //Az IE flag rs engedlyez jele
   .dbg_ie_din(dbg_data_in[4]),        //Az IE bitbe rand adat
   .dbg_reg_wr(dbg_reg_wr),            //A regisztertmb rs engedlyez jele
   .dbg_mem_wr(dbg_mem_wr),            //Az adatmemria rs engedlyez jele
   .dbg_mem_rd(dbg_mem_rd),            //Az adatmemria olvass engedlyez jele
   .dbg_is_brk(dbg_is_brk)             //A trspont llapot jelzse
);


endmodule
