//******************************************************************************
//* MiniRISC CPU v2.0                                                          *
//*                                                                            *
//* Utasts tpusok:                                                          *
//*            |15....12|11..........8|7......4|3...........0|                 *
//* -A tpus: | opkd  | rX/vezrls |   8 bites konstans   |                 *
//* -B tpus: |  1111  | rX/vezrls | opkd  | rY/vezrls |                 *
//*                                                                            *
//* A 4'b1111 prefix jelzi, hogy a msodik operandus nem konstans, hanem       *
//* regiszter, teht B tpus utastsrl van sz.                             *
//******************************************************************************
localparam REG_OP_PREFIX = 4'b1111;


//******************************************************************************
//* MOV rX, addr                                                      A tpus *
//* Adatmemria olvass abszolt cmzssel: rX <- DMEM[addr]          - - - -  *
//*                                                                            *
//*  |15..12|11.....8|7................0|                                      *
//*  | 1101 |   rX   | adatmemria cm  |                                      *
//*                                                                            *
//* MOV rX, (rY)                                                      B tpus *
//* Adatmemria olvass indirekt cmzssel: rX <- DMEM[rY]            - - - -  *
//*                                                                            *
//*  |15..12|11.....8|7......4|3.......0|                                      *
//*  | 1111 |   rX   |  1101  |    rY   |                                      *
//******************************************************************************
localparam OPCODE_LD = 4'b1101;


//******************************************************************************
//* MOV addr, rX                                                      A tpus *
//* Adatmemria rs abszolt cmzssel: DMEM[addr] <- rX             - - - -  *
//*                                                                            *
//*  |15..12|11.....8|7................0|                                      *
//*  | 1001 |   rX   | adatmemria cm  |                                      *
//*                                                                            *
//* MOV (rY), rX                                                      B tpus *
//* Adatmemria rs indirekt cmzssel: DMEM[rY] <- rX               - - - -  *
//*                                                                            *
//*  |15..12|11.....8|7......4|3.......0|                                      *
//*  | 1111 |   rX   |  1001  |    rY   |                                      *
//******************************************************************************
localparam OPCODE_ST = 4'b1001;


//******************************************************************************
//* MOV rX, #imm                                                      A tpus *
//* Konstans betltse regiszterbe: rX <- imm                         - - - -  *
//*                                                                            *
//*  |15..12|11.....8|7................0|                                      *
//*  | 1100 |   rX   | 8 bites konstans |                                      *
//*                                                                            *
//* MOV rX, rY                                                        B tpus *
//* Adatmozgats regiszterbl regiszterbe: rX <- rY                   - - - -  *
//*                                                                            *
//*  |15..12|11.....8|7......4|3.......0|                                      *
//*  | 1111 |   rX   |  1100  |    rY   |                                      *
//******************************************************************************
localparam OPCODE_MOV = 4'b1100;


//******************************************************************************
//* ADD rX, #imm                                                      A tpus *
//* Konstans hozzadsa regiszterhez tvitel nlkl: rX <- rX + imm   Z C N V  *
//*                                                                            *
//* ADC rX, #imm                                                      A tpus *
//* Konstans hozzadsa regiszterhez tvitellel: rX <- rX + imm + C   Z C N V  *
//*                                                                            *
//* SUB rX, #imm                                                      A tpus *
//* Konstans kivonsa regiszterbl tvitel nlkl: rX <- rX - imm     Z C N V  *
//*                                                                            *
//* SBC rX, #imm                                                      A tpus *
//* Konstans kivonsa regiszterbl tvitellel: rX <- rX - imm + C     Z C N V  *
//*                                                                            *
//*  |15..12|11.....8|7................0|                                      *
//*  | 00SC |   rX   | 8 bites konstans |                                      *
//*                                                                            *
//* ADD rX, rY                                                        B tpus *
//* Regiszter hozzadsa regiszterhez tvitel nlkl: rX <- rX + rY   Z C N V  *
//*                                                                            *
//* ADC rX, rY                                                        B tpus *
//* Regiszter hozzadsa regiszterhez tvitellel: rX <- rX + rY + C   Z C N V  *
//*                                                                            *
//* SUB rX, rY                                                        B tpus *
//* Regiszter kivonsa regiszterbl tvitel nlkl: rX <- rX - rY     Z C N V  *
//*                                                                            *
//* SBC rX, rY                                                        B tpus *
//* Regiszter kivonsa regiszterbl tvitellel: rX <- rX - rY + C     Z C N V  *
//*                                                                            *
//*  |15..12|11.....8|7......4|3.......0|                                      *
//*  | 1111 |   rX   |  00SC  |    rY   |                                      *
//*                                                                            *
//*  S: mvelet kivlasztsa (0: sszeads, 1: kivons)                        *
//*  C: tvitel kivlasztsa (0: tvitel nlkl, 1: tvitellel)                *
//******************************************************************************
localparam OPCODE_ADD = 4'b0000;
localparam OPCODE_ADC = 4'b0001;
localparam OPCODE_SUB = 4'b0010;
localparam OPCODE_SBC = 4'b0011;


//******************************************************************************
//* CMP rX, #imm                                                      A tpus *
//* Regiszter sszehasonltsa konstanssal: rX - imm                  Z C N V  *
//*                                                                            *
//*  |15..12|11.....8|7................0|                                      *
//*  | 1010 |   rX   | 8 bites konstans |                                      *
//*                                                                            *
//* CMP rX, rY                                                        B tpus *
//* Regiszter sszehasonltsa regiszterrel: rX - rY                  Z C N V  *
//*                                                                            *
//*  |15..12|11.....8|7......4|3.......0|                                      *
//*  | 1111 |   rX   |  1010  |    rY   |                                      *
//******************************************************************************
localparam OPCODE_CMP = 4'b1010;


//******************************************************************************
//* AND rX, #imm                                                      A tpus *
//* Bitenknti S konstanssal: rX <- rX & imm                         Z - N -  *
//*                                                                            *
//* OR  rX, #imm                                                      A tpus *
//* Bitenknti VAGY konstanssal: rX <- rX | imm                       Z - N -  *
//*                                                                            *
//* XOR rX, #imm                                                      A tpus *
//* Bitenknti XOR konstanssal: rX <- rX ^ imm                        Z - N -  *
//*                                                                            *
//* SWP rX                                                            A tpus *
//* Als/fels 4 bit felcserlse: rX <- {rX[3:0], rX[7:4]}           Z - N -  *
//*                                                                            *
//*  |15..12|11.....8|7................0|                                      *
//*  | 01AB |   rX   | 8 bites konstans |                                      *
//*                                                                            *
//* AND rX, rY                                                        B tpus *
//* Bitenknti S regiszterrel: rX <- rX & rY                         Z - N -  *
//*                                                                            *
//* OR  rX, rY                                                        B tpus *
//* Bitenknti VAGY regiszterrel: rX <- rX | rY                       Z - N -  *
//*                                                                            *
//* XOR rX, rY                                                        B tpus *
//* Bitenknti XOR regiszterrel: rX <- rX ^ rY                        Z - N -  *
//*                                                                            *
//*  |15..12|11.....8|7......4|3.......0|                                      *
//*  | 1111 |   rX   |  01AB  |    rY   |                                      *
//*                                                                            *
//*  AB: mvelet kivlasztsa (00: S, 01: VAGY, 10: XOR, 11: csere)           *
//*                                                                            *
//*  Megjegyzs: a csere mvelet csak A tpus utastsnl rtelmezett,        *
//*  B tpus utastsnl ez a mveleti kd shiftelst/forgatst hajt vgre.   *
//******************************************************************************
localparam OPCODE_AND = 4'b0100;
localparam OPCODE_OR  = 4'b0101;
localparam OPCODE_XOR = 4'b0110;

localparam LOGIC_AND = 2'b00;
localparam LOGIC_OR  = 2'b01;
localparam LOGIC_XOR = 2'b10;


//******************************************************************************
//* TST rX, #imm                                                      A tpus *
//* Bittesztels konstanssal: rX & imm                                Z - N -  *
//*                                                                            *
//*  |15..12|11.....8|7................0|                                      *
//*  | 1000 |   rX   | 8 bites konstans |                                      *
//*                                                                            *
//* TST rX, rY                                                        B tpus *
//* Bittesztels regiszterrel: rX & rY                                Z - N -  *
//*                                                                            *
//*  |15..12|11.....8|7......4|3.......0|                                      *
//*  | 1111 |   rX   |  1000  |    rY   |                                      *
//******************************************************************************
localparam OPCODE_TST = 4'b1000;


//******************************************************************************
//* SL0 rX                                                            B tpus *
//* Shiftels balra (0): rX <- {rX[6:0], 0}, C <- rX[7]               Z C N -  *
//*                                                                            *
//* SL1 rX                                                            B tpus *
//* Shiftels balra (1): rX <- {rX[6:0], 1}, C <- rX[7]               Z C N -  *
//*                                                                            *
//* SR0 rX                                                            B tpus *
//* Shiftels jobbra (0): rX <- {0, rX[7:1]}, C <- rX[0]              Z C N -  *
//*                                                                            *
//* SR1 rX                                                            B tpus *
//* Shiftels jobbra (1): rX <- {1, rX[7:1]}, C <- rX[0]              Z C N -  *
//*                                                                            *
//* ASR rX                                                            B tpus *
//* Aritmetikai shift jobbra: rX <- {rX[7], rX[7:1]}, C <- rX[0]      Z C N -  *
//*                                                                            *
//* ROL rX                                                            B tpus *
//* Forgats balra: rX <- {rX[6:0], rX[7]}, C <- rX[7]                Z C N -  *
//*                                                                            *
//* ROR rX                                                            B tpus *
//* Forgats jobbra: rX <- {rX[0], rX[7:1]}, C <- rX[0]               Z C N -  *
//*                                                                            *
//* RLC rX                                                            B tpus *
//* Forgats balra carry-vel: rX <- {rX[6:0], C}, C <- rX[7]          Z C N -  *
//*                                                                            *
//* RRC rX                                                            B tpus *
//* Forgats jobbra carry-vel: rX <- {C, rX[7:1]}, C <- rX[0]         Z C N -  *
//*                                                                            *
//*  |15..12|11.....8|7......4|3.......0|                                      *
//*  | 1111 |   rX   |  0111  |  AIRD   |                                      *
//*                                                                            *
//*  D: irny kivlasztsa (0: balra, 1: jobbra)                               *
//*  R: mvelet kivlasztsa (0: shiftels, 1: forgats)                       *
//*  I: a beshiftelt bit rtke/kivlasztsa (0: 0/kishiftelt bit, 1: 1/carry) *
//*  A: a shiftels tpusa (0: norml, 1: aritmetikai)                         *
//******************************************************************************
localparam OPCODE_SHIFT = 4'b0111;

localparam SHIFT_SHL  = 2'b00;
localparam SHIFT_SHR  = 2'b01;
localparam SHIFT_ROL  = 2'b10;
localparam SHIFT_ROR  = 2'b11;


//******************************************************************************
//* JMP addr - Felttel nlkli ugrs (PC <- addr)                    A tpus *
//* JZ  addr - Ugrs, ha a Z flag 1   (PC <- addr, ha Z=1)            - - - -  *
//* JNZ addr - Ugrs, ha a Z flag 0   (PC <- addr, ha Z=0)                     *
//* JC  addr - Ugrs, ha a C flag 1   (PC <- addr, ha C=1)                     *
//* JNC addr - Ugrs, ha a C flag 0   (PC <- addr, ha C=0)                     *
//* JN  addr - Ugrs, ha az N flag 1  (PC <- addr, ha N=1)                     *
//* JNN addr - Ugrs, ha az N flag 0  (PC <- addr, ha N=0)                     *
//* JV  addr - Ugrs, ha a V flag 1   (PC <- addr, ha V=1)                     *
//* JNV addr - Ugrs, ha a V flag 0   (PC <- addr, ha V=0)                     *
//* JSR addr - Szubrutinhvs         (stack <- PC <- addr)                    *
//*                                                                            *
//*  |15..12|11.....8|7................0|                                      *
//*  | 1011 | mvelet|programmemria cm|                                      *
//*                                                                            *
//* JMP (rY) - Felttel nlkli ugrs (PC <- rY)                      B tpus *
//* JZ  (rY) - Ugrs, ha a Z flag 1   (PC <- rY, ha Z=1)              - - - -  *
//* JNZ (rY) - Ugrs, ha a Z flag 0   (PC <- rY, ha Z=0)                       *
//* JC  (rY) - Ugrs, ha a C flag 1   (PC <- rY, ha C=1)                       *
//* JNC (rY) - Ugrs, ha a C flag 0   (PC <- rY, ha C=0)                       *
//* JN  (rY) - Ugrs, ha az N flag 1  (PC <- rY, ha N=1)                       *
//* JNN (rY) - Ugrs, ha az N flag 0  (PC <- rY, ha N=0)                       *
//* JV  (rY) - Ugrs, ha a V flag 1   (PC <- rY, ha V=1)                       *
//* JNV (rY) - Ugrs, ha a V flag 0   (PC <- rY, ha V=0)                       *
//* JSR (rY) - Szubrutinhvs         (stack <- PC <- rY)                      *
//*                                                                            *
//*  |15..12|11.....8|7......4|3.......0|                                      *
//*  | 1111 | mvelet|  1011  |    rY   |                                      *
//*                                                                            *
//* RTS - Visszatrs szubrutinbl    (PC <- stack)                   A tpus *
//* RTI - Visszatrs megszaktsbl  (PC,Z,C,N,V,IE <- stack)        - - - -  *
//* CLI - Megszaktsok tiltsa       (IE <- 0)                                *
//* STI - Megszaktsok engedlyezse (IE <- 1)                                *
//*                                                                            *
//*  |15..12|11.....8|7................0|                                      *
//*  | 1011 | mvelet|     00000000     |                                      *
//******************************************************************************
localparam OPCODE_CTRL = 4'b1011;

localparam CTRL_JMP = 4'b0000;
localparam CTRL_JZ  = 4'b0001;
localparam CTRL_JNZ = 4'b0010;
localparam CTRL_JC  = 4'b0011;
localparam CTRL_JNC = 4'b0100;
localparam CTRL_JN  = 4'b0101;
localparam CTRL_JNN = 4'b0110;
localparam CTRL_JV  = 4'b0111;
localparam CTRL_JNV = 4'b1000;
localparam CTRL_JSR = 4'b1001;
localparam CTRL_RTS = 4'b1010;
localparam CTRL_RTI = 4'b1011;
localparam CTRL_CLI = 4'b1100;
localparam CTRL_STI = 4'b1101;

