`timescale 1ns / 1ps

//******************************************************************************
//* Egyszer DMA vezrl.                                                      *
//*                                                                            *
//* A DMA tvitel a forrs- s clcmek megadsa utn az adatmret regiszter   *
//* rsval indthat. A cmek automatikus nvelse engedlyezhet, illetve   *
//* tilthat. Az adattvitel vgt megszakts jelzi, ha engedlyezve van.     *
//*                                                                            *
//* A DMA vezrl regisztereinek elrse a slave busz interfszen keresztl    *
//* trtnik, a DMA tvitel pedig a master busz interfszen keresztl.         *
//*                                                                            *
//* A perifria cme paramter tadssal llthat be a fels szint modulban  *
//* az megpldnyostskor. A szintzis mr ennek megfeleleen az aktulis     *
//* BASEADDR bziscmmel trtnik.                                             *
//*                                                                            *
//* A programozi fellet:                                                     *
//*                                                                            *
//* Cm         Tpus   Bitek                                                  *
//* BASEADDR+0  WR      A DMA vezrl parancs regisztere                       *
//*                     -     -     -     -     -     IEN   DINC  SINC         *
//* BASEADDR+0  RD      A DMA vezrl sttusz regisztere                       *
//*                     BUSY  IRQ   IFLG  0     0     IEN   DINC  SINC         *
//* BASEADDR+1  WR      Forrscm regiszter                                    *
//* BASEADDR+2  WR      Clcm regiszter                                       *
//* BASEADDR+3  WR      Adatmret regiszter                                    *
//*                                                                            *
//* A parancs regiszter bitjei statikus set/clear rtelemben hasznlhatk. A   *
//* sttusz regiszter md bitjei statikusak, az esemnyjelz IFLG (s ezzel    * 
//* egytt az esetleg aktv IRQ is) 1 bersra trldik.                      *
//******************************************************************************
module dma_controller #(
   //A perifria bziscme.
   parameter BASEADDR = 8'hff
) (
   //rajel s reset.
   input  wire       clk,              //rajel
   input  wire       rst,              //Reset jel
   
   //A slave busz interfsz jelei (regiszter elrs).
   input  wire [7:0] s_mst2slv_addr,   //Cmbusz
   input  wire       s_mst2slv_wr,     //rs engedlyez jel
   input  wire       s_mst2slv_rd,     //Olvass engedlyez jel
   input  wire [7:0] s_mst2slv_data,   //rsi adatbusz
   output wire [7:0] s_slv2mst_data,   //Olvassi adatbusz
   
   //A master busz interfsz jelei (DMA tvitel).
   output wire       m_bus_req,        //Busz hozzfrs krse
   input  wire       m_bus_grant,      //Busz hozzfrs megadsa
   output reg  [7:0] m_mst2slv_addr,   //Cmbusz
   output wire       m_mst2slv_wr,     //rs engedlyez jel
   output wire       m_mst2slv_rd,     //Olvass engedlyez jel
   output wire [7:0] m_mst2slv_data,   //rsi adatbusz
   input  wire [7:0] m_slv2mst_data,   //Olvassi adatbusz
   
   //Megszaktskr kimenet.
   output wire       irq
);

//******************************************************************************
//* Cmdekdols.                                                              *
//******************************************************************************
//A perifria kivlaszt jele.
wire psel = ((s_mst2slv_addr >> 2) == (BASEADDR >> 2));

//A parancs regiszter rsnak jelzse.
wire cmd_wr   = psel & s_mst2slv_wr & (s_mst2slv_addr[1:0] == 2'b00); 

//A sttusz regiszter olvassnak jelzse.
wire stat_rd  = psel & s_mst2slv_rd & (s_mst2slv_addr[1:0] == 2'b00);

//A forrscm regiszter rsnak jelzse. 
wire saddr_wr = psel & s_mst2slv_wr & (s_mst2slv_addr[1:0] == 2'b01);

//A clcm regiszter rsnak jelzse. 
wire daddr_wr = psel & s_mst2slv_wr & (s_mst2slv_addr[1:0] == 2'b10);

//Az adatmret regiszter rsnak jelzse. 
wire len_wr   = psel & s_mst2slv_wr & (s_mst2slv_addr[1:0] == 2'b11);


//******************************************************************************
//* A DMA vezrl parancs s sttusz regisztere.                               *
//******************************************************************************
wire busy;
wire done;

//A parancs regiszter SINC bitje (forrscm nvels engedlyezse).
reg sinc;

always @(posedge clk)
begin
   if (rst)
      sinc <= 1'b1;
   else
      if (cmd_wr && (busy == 0))
         sinc <= s_mst2slv_data[0];
end

//A parancs regiszter DINC bitje (clcm nvels engedlyezse).
reg dinc;

always @(posedge clk)
begin
   if (rst)
      dinc <= 1'b1;
   else
      if (cmd_wr && (busy == 0))
         dinc <= s_mst2slv_data[1];
end

//A parancs regiszter IEN bitje (megszakts engedlyezs).
reg ien;

always @(posedge clk)
begin
   if (rst)
      ien <= 1'b0;
   else
      if (cmd_wr && (busy == 0))
         ien <= s_mst2slv_data[2];
end

//A sttusz regiszter IFLG bitje (megszakts flag). Belltdik, ha
//vget rt a DMA adattvitel. A jelzs 1 bersval trlhet.
reg iflg;

always @(posedge clk)
begin
   if (rst)
      iflg <= 1'b0;
   else
      if (done)
         iflg <= 1'b1;
      else
         if (cmd_wr && (busy == 0) && s_mst2slv_data[5])
            iflg <= 1'b0;
end

//A 8 bites sttusz informci sszelltsa az egyedi bitek alapjn.
//Az IRQ s az IFLG bit majdnem azonos jelents, de az IRQ bit magba
//foglalja az IEN llapott is. 
wire [7:0] status_reg = {busy, irq, iflg, 2'b00, ien, dinc, sinc};

//A slave olvassi adatbusz meghajtsa. Csak a sttusz regiszter olvashat.
assign s_slv2mst_data = (stat_rd) ? status_reg : 8'd0;

//A megszaktskr kimenet meghajtsa.
assign irq = ien & iflg;


//******************************************************************************
//* A forrscm szmll.                                                      *
//******************************************************************************
reg  [7:0] saddr_cnt;
wire       saddr_cnt_en;

always @(posedge clk)
begin
   if (saddr_wr && (busy == 0))
      saddr_cnt <= s_mst2slv_data;
   else
      if (saddr_cnt_en && sinc)
         saddr_cnt <= saddr_cnt + 8'd1;
end


//******************************************************************************
//* A clcm szmll.                                                         *
//******************************************************************************
reg  [7:0] daddr_cnt;
wire       daddr_cnt_en;

always @(posedge clk)
begin
   if (daddr_wr && (busy == 0))
      daddr_cnt <= s_mst2slv_data;
   else
      if (daddr_cnt_en && dinc)
         daddr_cnt <= daddr_cnt + 8'd1;
end


//******************************************************************************
//* Az adatmret szmll.                                                     *
//******************************************************************************
reg  [7:0] len_cnt;
wire       len_cnt_en;
wire       len_cnt_tc = (len_cnt == 8'd0);

always @(posedge clk)
begin
   if (len_wr && (busy == 0))
      len_cnt <= s_mst2slv_data;
   else
      if (len_cnt_en)
         len_cnt <= len_cnt - 8'd1;
end


//******************************************************************************
//* A vezrl llapotgp.                                                      *
//******************************************************************************
localparam STATE_IDLE = 2'd0;
localparam STATE_RD   = 2'd1;
localparam STATE_WR   = 2'd2;
localparam STATE_DONE = 2'd3;

reg [1:0] state;

always @(posedge clk)
begin
   if (rst)
      state <= STATE_IDLE;
   else
      case (state)
         //Vrakozs a DMA adattvitel indtsra. Az adatmret
         //regiszter rsval indthat az tvitel.
         STATE_IDLE: if (len_wr)
                        state <= STATE_RD;
                     else
                        state <= STATE_IDLE;
                        
         //Memria olvass a forrscmrl.
         STATE_RD  : if (m_bus_grant)
                        state <= STATE_WR;
                     else
                        state <= STATE_RD;
         
         //Memria rs a clcmre s a htralv bjtok szmnak
         //vizsglata.
         STATE_WR  : if (m_bus_grant)
                        if (len_cnt_tc)
                           state <= STATE_DONE;
                        else
                           state <= STATE_RD;
                     else
                        state <= STATE_WR;
                        
         //A DMA tvitel befejezdsnek jelzse.
         STATE_DONE: state <= STATE_IDLE;
      endcase
end

//Foglaltsg jelzs.
assign busy = ~(state == STATE_IDLE);

//Az adattvitel befejezdsnek jelzse.
assign done = (state == STATE_DONE);

//A szmllk engedlyez jelei.
assign saddr_cnt_en = (state == STATE_RD) & m_bus_grant;
assign daddr_cnt_en = (state == STATE_WR) & m_bus_grant;
assign len_cnt_en   = (state == STATE_RD) & m_bus_grant;


//******************************************************************************
//* A master busz interfsz kimeneteinek meghajtsa.                           *
//******************************************************************************
//Busz hozzfrs krse.
assign m_bus_req = (state == STATE_RD) | (state == STATE_WR);

//A cmbuszra memria olvass esetn a forrscm, memria rs
//esetn a clcm, egybknt pedig inaktv nulla rtk kerl.
always @(*)
begin
   if (m_bus_grant)
      case (state)
         STATE_RD: m_mst2slv_addr <= saddr_cnt;
         STATE_WR: m_mst2slv_addr <= daddr_cnt;
         default : m_mst2slv_addr <= 8'd0;
      endcase
   else
      m_mst2slv_addr <= 8'd0;
end

//rs s olvass engedlyez jelek.
assign m_mst2slv_wr = (state == STATE_WR) & m_bus_grant;
assign m_mst2slv_rd = (state == STATE_RD) & m_bus_grant;

//A slave egysgrl beolvasott adatot trol regiszter.
reg [7:0] data_reg;

always @(posedge clk)
begin
   if (m_mst2slv_rd)
      data_reg <= m_slv2mst_data;
end

//Az rsi adatbusz meghajtsa.
assign m_mst2slv_data = (m_mst2slv_wr) ? data_reg : 8'd0;

endmodule
