`timescale 1ns / 1ps

//******************************************************************************
//* Egyszer 8 bites bemeneti modul megszaktskrsi lehetsggel a bemeneti  * 
//* jelek megvltozsa esetn. A gpio_in bemenet 200 Hz-es mintavtelezssel   *
//* pergsmentestve van.                                                      *
//*                                                                            *
//* A perifria cme paramter tadssal llthat be a fels szint modulban. *
//* a GPIO modul megpldnyostsakor. A szintzis mr ennek megfeleleen az   *
//* aktulis BASEADDR bziscmmel trtnik.                                    *
//*                                                                            *
//* A programozi fellet:                                                     *
//*                                                                            *
//* Cm         Tpus   Bitek                                                  *
//* BASEADDR+0  RD      Adatregiszter                                          *
//*                     IN7   IN6   IN5   IN4   IN3   IN2   IN1   IN0          *
//* BASEADDR+1  RD/WR   Megszakts engedlyez regiszter                      *
//*                     IE7   IE6   IE5   IE4   IE3   IE2   IE1   IE0          *
//* BASEADDR+2  RD/W1C  Megszakts flag regiszter                             *
//*                     IF7   IF6   IF5   IF4   IF3   IF2   IF1   IF0          *
//*                                                                            *
//* A megszakts flag regiszter bitjei 1 bersval trlhetek, ezzel        *
//* nyugtzva a megszaktskrst.                                             * 
//******************************************************************************
module basic_in_irq #(
   //A perifria bziscme.
   parameter BASEADDR = 8'hff
) (
   //rajel s reset.
   input  wire       clk,              //rajel
   input  wire       rst,              //Reset jel
   
   //A slave busz interfsz jelei.
   input  wire [7:0] s_mst2slv_addr,   //Cmbusz
   input  wire       s_mst2slv_wr,     //rs engedlyez jel
   input  wire       s_mst2slv_rd,     //Olvass engedlyez jel
   input  wire [7:0] s_mst2slv_data,   //rsi adatbusz
   output reg  [7:0] s_slv2mst_data,   //Olvassi adatbusz
   
   //Megszaktskr kimenet.
   output wire       irq,
   
   //A GPIO interfsz jelei.
   input  wire [7:0] gpio_in           //Az IO lbak aktulis rtke
);

//******************************************************************************
//* Cmdekdols.                                                              *
//******************************************************************************
//A perifria kivlaszt jele.
wire psel = ((s_mst2slv_addr >> 2) == (BASEADDR >> 2));

//Az adatregiszter olvassnak jelzse.
wire in_reg_rd = psel & s_mst2slv_rd & (s_mst2slv_addr[1:0] == 2'b00);

//A megszakts engedlyez regiszter rsnak s olvassnak jelzse.
wire ie_reg_wr = psel & s_mst2slv_wr & (s_mst2slv_addr[1:0] == 2'b01);
wire ie_reg_rd = psel & s_mst2slv_rd & (s_mst2slv_addr[1:0] == 2'b01);

//A megszakts flag regiszter rsnak s olvassnak jelzse.
wire if_reg_wr = psel & s_mst2slv_wr & (s_mst2slv_addr[1:0] == 2'b10);
wire if_reg_rd = psel & s_mst2slv_rd & (s_mst2slv_addr[1:0] == 2'b10);


//******************************************************************************
//* A 200 Hz-es engedlyez jel ellltsa a bemenet pergsmentestshez.    *
//* A szmll modulusa: 16000000 Hz / 200 Hz = 80000 (79999-0).               * 
//******************************************************************************
reg  [16:0] clk_div_cnt;
wire        gpio_in_sample = (clk_div_cnt == 0);

always @(posedge clk)
begin
   if (rst || gpio_in_sample)
      clk_div_cnt <= 17'd79999;
   else
      clk_div_cnt <= clk_div_cnt - 17'd1;
end


//******************************************************************************
//* Az adatregiszter.                                                          *
//******************************************************************************
reg [7:0] in_reg;

always @(posedge clk)
begin
   if (rst)
      in_reg <= 8'd0;                  //Reset esetn trljk a regisztert
   else
      if (gpio_in_sample)              //Egybknt 200 Hz-el folyamatosan
         in_reg <= gpio_in;            //mintavtelezzk a bemenetet
end

//A bemeneti vltozs detektlsa.
reg  [7:0] in_reg_prev;
wire [7:0] in_reg_changed = in_reg ^ in_reg_prev;

always @(posedge clk)
begin
   if (rst)
      in_reg_prev <= 8'd0;
   else
      in_reg_prev <= in_reg;
end


//******************************************************************************
//* A megszakts engedlyez regiszter.                                       *
//******************************************************************************
reg [7:0] ie_reg;

always @(posedge clk)
begin
   if (rst)
      ie_reg <= 8'd0;                  //Reset: tiltjuk a megszaktsokat
   else
      if (ie_reg_wr)
         ie_reg <= s_mst2slv_data;     //Regiszter rs
end


//******************************************************************************
//* A megszakts flag regiszter.                                              *
//******************************************************************************
reg [7:0] if_reg;

integer i;

always @(posedge clk)
begin
   for (i = 0; i < 8; i = i + 1)
      if (rst)
         if_reg[i] <= 1'b0;            //Reset: a jelzsek trlse
      else
         if (in_reg_changed[i])
            if_reg[i] <= 1'b1;         //A bemenet megvltozsnak jelzse
         else
            if (if_reg_wr && s_mst2slv_data[i])
               if_reg[i] <= 1'b0;      //1 bersa esetn trljk a jelzst
end

//A megszaktskr kimenet meghajtsa.
assign irq = |(if_reg & ie_reg);


//******************************************************************************
//* Az olvassi adatbusz meghajtsa.                                           *
//******************************************************************************
wire [2:0] dout_sel = {if_reg_rd, ie_reg_rd, in_reg_rd};

always @(*)
begin
   case (dout_sel)
      3'b001 : s_slv2mst_data <= in_reg;
      3'b010 : s_slv2mst_data <= ie_reg;
      3'b100 : s_slv2mst_data <= if_reg;
      default: s_slv2mst_data <= 8'd0;
   endcase
end

endmodule
