`timescale 1ns / 1ps

//******************************************************************************
//* Egyszer 8 bites ktirny I/O modul.                                      *
//*                                                                            *
//* A perifria cme paramter tadssal llthat be a fels szint modulban  *
//* a GPIO modul megpldnyostsakor. A szintzis mr ennek megfeleleen az   *
//* aktulis BASEADDR bziscmmel trtnik.                                    *
//*                                                                            *
//* A BASEADDR cm szolgl a kimeneti regiszter rsra s visszaolvassra,   *
//* a BASEADDR+1 cmrl olvashat be a GPIO lbak aktulis rtke, valamint    *
//* a BASEADDR+2 cm szolgl az irny regiszter rsra s visszaolvassra.   * 
//*                                                                            *
//* Teht a BASIC_IO perifria a szoksos mikrovezrlk port GPIO funkcijt   *
//* valstja meg. Termszetesen egy adott idben egy GPIO vonal vagy csak     * 
//* kimenet, vagy csak bemenet lehet, mg ha ezt a programbl dinamikusan      * 
//* llthatjuk is. St klns figyelmet ignyel egy ramkrn bell az adott *
//* vonal kls interfsze, elkerlend az ellenttes llapot kimenetek       *
//* sszekapcsolst!                                                          *
//*                                                                            *
//* Egyetlen IC lb (=Port pin) funkcija grafikusan brzolva:                *
//*                                                                            *
//*                   -------               Vcc                                *
//*    DIR         >--|D   Q|---------|     |                                  *
//*                   |     |  |      |     |                                  *
//*                   |>    |  |      |    |-|  Rf felhz ellenlls          *
//*                   -------  |      |    | |                                 *
//*    DIR olvass <-----------|    |\|    |-|                                 *
//*                   -------       | \     |                                  *
//*    adat KI     >--|D   Q|-------|  >-------------<-> az IC programozhat   *
//*                   |     |  |    | /        |         IO lba               *
//*                   |>    |  |    |/         |                               *
//*                   -------  |               |                               *
//*    KI olvass  <-----------|               |                               *
//*                   -------                  |                               *
//*    IO lb olv. <--|Q   D|------------------|                               *
//*                   |     |                                                  *
//*                   |    <|                                                  *
//*                   -------                                                  *
//*                                                                            *
//* A programozi fellet:                                                     *
//*                                                                            *
//* Cm         Tpus   Bitek                                                  *
//* BASEADDR+0  RD/WR   Kimeneti adatregiszter                                 *
//*                     OUT7  OUT6  OUT5  OUT4  OUT3  OUT2  OUT1  OUT0         *
//* BASEADDR+1  RD      Adat az IO lbakon                                     *  
//*                     IN7   IN6   IN5   IN4   IN3   IN2   IN1   IN0          *
//* BASEADDR+2  RD/WR   Az IO lb irnynak belltsa (0: be, 1:ki)           *
//*                     DIR7  DIR6  DIR5  DIR4  DIR3  DIR2  DIR1  DIR0         *
//* BASEADDR+3  RD      Nem hasznlt, olvass esetn mindig 0                  *
//*                                                                            *
//* A processzor rsi (WRITE) ciklusnak iddiagramja:                        *
//* Az rsi ciklust az 1 rajel ciklus ideig aktv S_MST2SLV_WR jel jelzi. Az *
//* rsi ciklus ideje alatt a cm s a kimeneti adat stabil.                  *
//*                                                                            *
//*                     --------          --------          --------           *
//* CLK                |        |        |        |        |        |          *
//*                ----          --------          --------          ----      *
//*                                        ------------------                  *
//* S_MST2SLV_WR                          /                  \                 *
//*                -----------------------                    -----------      *
//*                ----------------------- ------------------ -----------      *
//* S_MST2SLV_ADDR                        X   RVNYES CM   X                 *
//*                ----------------------- ------------------ -----------      *
//*                ----------------------- ------------------ -----------      *
//* S_MST2SLV_DATA                        X   RVNYES ADAT  X                 *
//*                ----------------------- ------------------ -----------      *
//*                                                        ^                   *
//*                                Az rsi parancs vgrehajtsa itt trtnik. *
//*                                                                            *
//* A processzor olvassi (READ) ciklusnak iddiagramja:                      *
//* Az olvassi ciklust az 1 rajel ciklus ideig aktv S_MST2SLV_RD jel jelzi. *
//* Az olvassi ciklus ideje alatt a cm stabil s a kivlasztott perifria az *
//* olvassi adatbuszra kapuzza az adatot. Az olvassi ciklus eltt s utn    *
//* az olvassi adatbusz rtke inaktv 0 kell, hogy legyen.                   *
//*                                                                            *
//*                     --------          --------          --------           *
//* CLK                |        |        |        |        |        |          *
//*                ----          --------          --------          ----      *
//*                                        ------------------                  *
//* S_MST2SLV_RD                          /                  \                 *
//*                -----------------------                    -----------      *
//*                ----------------------- ------------------ -----------      *
//* S_MST2SLV_ADDR                        X   RVNYES CM   X                 *
//*                ----------------------- ------------------ -----------      *
//*                ----------------------- ------------------ -----------      *
//* S_SLV2MST_DATA           0            X   RVNYES ADAT  X       0         *
//*                ----------------------- ------------------ -----------      *
//*                                                        ^                   *
//*                              A bemeneti adat mintavtelezse itt trtnik. *   
//******************************************************************************
module basic_io #(
   //A perifria bziscme.
   parameter BASEADDR = 8'hff
) (
   //rajel s reset.
   input  wire       clk,              //rajel
   input  wire       rst,              //Reset jel
   
   //A slave busz interfsz jelei.
   input  wire [7:0] s_mst2slv_addr,   //Cmbusz
   input  wire       s_mst2slv_wr,     //rs engedlyez jel
   input  wire       s_mst2slv_rd,     //Olvass engedlyez jel
   input  wire [7:0] s_mst2slv_data,   //rsi adatbusz
   output reg  [7:0] s_slv2mst_data,   //Olvassi adatbusz
   
   //A GPIO interfsz jelei.
   output wire [7:0] gpio_out,         //Az IO lbakra kirand adat
   input  wire [7:0] gpio_in,          //Az IO lbak aktulis rtke
   output wire [7:0] gpio_dir          //A kimeneti meghajt engedlyez jele
);
  
//******************************************************************************
//* Cmdekdols.                                                              *
//******************************************************************************
//A perifria kivlaszt jele.
wire psel = ((s_mst2slv_addr >> 2) == (BASEADDR >> 2));

//A kimeneti adatregiszter rsnak s olvassnak jelzse.
wire out_reg_wr = psel & s_mst2slv_wr & (s_mst2slv_addr[1:0] == 2'b00);
wire out_reg_rd = psel & s_mst2slv_rd & (s_mst2slv_addr[1:0] == 2'b00);

//A bemeneti adatregiszter olvassnak jelzse.
wire in_reg_rd  = psel & s_mst2slv_rd & (s_mst2slv_addr[1:0] == 2'b01);

//Az irny kivlaszt regiszter rsnak s olvassnak jelzse.
wire dir_reg_wr = psel & s_mst2slv_wr & (s_mst2slv_addr[1:0] == 2'b10);
wire dir_reg_rd = psel & s_mst2slv_rd & (s_mst2slv_addr[1:0] == 2'b10);


//******************************************************************************
//* A kimeneti adatregiszter.                                                  *
//******************************************************************************
reg [7:0] out_reg;

always @(posedge clk)
begin
   if (rst)
      out_reg <= 8'd0;                 //Reset esetn trljk a regisztert
   else
      if (out_reg_wr)
         out_reg <= s_mst2slv_data;    //Kimeneti adatregiszter rsa
end

//Az IO lbak meghajtsa.
assign gpio_out = out_reg;


//******************************************************************************
//* A bemeneti adatregiszter.                                                  *
//******************************************************************************
reg [7:0] in_reg;

always @(posedge clk)
begin
   if (rst)
      in_reg <= 8'd0;                  //Reset esetn trljk a regisztert
   else
      in_reg <= gpio_in;               //Egybknt folyamatosan mintavtelezzk
end                                    //az IO lbak rtkt


//******************************************************************************
//* Az irny kivlaszt regiszter.                                             *
//* 0: az adott IO lb irnya bemenet                                          *
//* 1: az adott IO lb irnya kimenet                                          *
//******************************************************************************
reg [7:0] dir_reg;

always @(posedge clk)
begin
   if (rst)
      dir_reg <= 8'd0;                 //Reset esetn minden IO lb bemenet
   else
      if (dir_reg_wr)
         dir_reg <= s_mst2slv_data;    //Irny kivlaszt regiszter rsa
end

//Az irny kivlaszt kimenet meghajtsa.
assign gpio_dir = dir_reg;


//******************************************************************************
//* A processzor olvassi adatbusznak meghajtsa. Az olvassi adatbuszra csak *
//* az olvass ideje alatt kapcsoljuk r a krt rtket, egybknt egy inaktv *
//* nulla rtk jelenik meg rajta (elosztott busz multiplexer funkci).        *
//******************************************************************************
wire [2:0] dout_sel = {dir_reg_rd, in_reg_rd, out_reg_rd};

always @(*)
begin
   case (dout_sel)
      3'b001 : s_slv2mst_data <= out_reg;    //A kimeneti adatregiszter olvassa
      3'b010 : s_slv2mst_data <= in_reg;     //A bemeneti adatregiszter olvassa
      3'b100 : s_slv2mst_data <= dir_reg;    //Az irny kivlaszt reg. olvassa
      default: s_slv2mst_data <= 8'd0;       //Egybknt inaktv nulla rtk
   endcase
end

endmodule
