`timescale 1ns / 1ps

//******************************************************************************
//* Soros kommunikcit biztost slave USRT perifria.                        *
//*                                                                            *
//* Az USRT kommunikci keretezett formtumot hasznl: 1 START bit (0),       *
//* 8 adatbit (LSb elszr) s 1 STOP bit (1), a soros adatvonal inaktv       *
//* szintje 1. Az USRT rajel hatroza meg az adattviteli sebessget, az      *
//* rajelet a master egysg adja ki a slave egysg fel. Az ad az USRT       *
//* rajel felfut lre adja ki a kvetkez bitet, melyet a vev a lefut     *
//* lre mintavtelez. A vev csak a kerethiba mentes karaktereket trolja     *
//* (STOP bit = 1), a hibs karaktereket eldobja.                              *
//*                                                                            *
//* A perifria cme paramter tadssal llthat be a fels szint modulban  *
//* az USRT modul megpldnyostsakor. A szintzis mr ennek megfeleleen az  *
//* aktulis BASEADDR bziscmmel trtnik.                                    *
//*                                                                            *
//* A programozi fellet:                                                     *
//*                                                                            *
//* Cm         Tpus   Bitek                                                  *
//* BASEADDR+0          Kontroll regiszter                                     *
//*             WR      -     -     -     -     RXCLR TXCLR RXEN  TXEN         *
//*             RD      0     0     0     0     0     0     RXEN  TXEN         *
//* BASEADDR+1          FIFO sttusz regiszter                                 *
//*             RD      0     0     0     0     RXFUL RXNE  TXNF  TXEMP        *
//* BASEADDR+2          Megszakts engedlyez regiszter                      *
//*             WR      -     -     -     -     RXFUL RXNE  TXNF  TXEMP        *
//*             RD      0     0     0     0     RXFUL RXNE  TXNF  TXEMP        *
//* BASEADDR+3  WR      Az adsi FIFO rsa (ha TXNF=1)                        *
//* BASEADDR+3  RD      A vteli FIFO olvassa (ha RXNE=1)                     *
//*                     D7    D6    D5    D4    D3    D2    D1    D0           *
//******************************************************************************
module slave_usrt #(
   //A perifria bziscme.
   parameter BASEADDR = 8'hff
) (
   //rajel s reset.
   input  wire       clk,              //rajel
   input  wire       rst,              //Reset jel
   
   //A soros interfsz jelei.
   input  wire       usrt_clk,         //USRT rajel
   input  wire       usrt_rxd,         //Soros adatbemenet
   output wire       usrt_txd,         //Soros adatkimenet
   
   //A slave busz interfsz jelei.
   input  wire [7:0] s_mst2slv_addr,   //Cmbusz
   input  wire       s_mst2slv_wr,     //rs engedlyez jel
   input  wire       s_mst2slv_rd,     //Olvass engedlyez jel
   input  wire [7:0] s_mst2slv_data,   //rsi adatbusz
   output reg  [7:0] s_slv2mst_data,   //Olvassi adatbusz
   
   //Megszaktskr kimenet.
   output wire       irq
);

//******************************************************************************
//* Cmdekdols.                                                              *
//******************************************************************************
//A perifria kivlaszt jele.
wire psel = ((s_mst2slv_addr >> 2) == (BASEADDR >> 2));

//A kontroll regiszter rsnak s olvassnak jelzse.
wire ctrl_reg_wr = psel & s_mst2slv_wr & (s_mst2slv_addr[1:0] == 2'b00);
wire ctrl_reg_rd = psel & s_mst2slv_rd & (s_mst2slv_addr[1:0] == 2'b00);

//A FIFO sttusz regiszter olvassnak jelzse.
wire stat_reg_rd = psel & s_mst2slv_rd & (s_mst2slv_addr[1:0] == 2'b01);

//A megszakts engedlyez regiszter rsnak s olvassnak jelzse.
wire ie_reg_wr   = psel & s_mst2slv_wr & (s_mst2slv_addr[1:0] == 2'b10);
wire ie_reg_rd   = psel & s_mst2slv_rd & (s_mst2slv_addr[1:0] == 2'b10);

//Az adatregiszter rsnak s olvassnak jelzse.
wire data_reg_wr = psel & s_mst2slv_wr & (s_mst2slv_addr[1:0] == 2'b11);
wire data_reg_rd = psel & s_mst2slv_rd & (s_mst2slv_addr[1:0] == 2'b11);


//******************************************************************************
//* A kontroll regiszter.                                                      *
//******************************************************************************
//A kontroll regiszter TXEN bitje (ads engedlyezs).
reg tx_enable;

always @(posedge clk)
begin
   if (rst)
      tx_enable <= 1'b0;
   else
      if (ctrl_reg_wr)
         tx_enable <= s_mst2slv_data[0];
end

//A kontroll regiszter RXEN bitje (vtel engedlyezs).
reg rx_enable;

always @(posedge clk)
begin
   if (rst)
      rx_enable <= 1'b0;
   else
      if (ctrl_reg_wr)
         rx_enable <= s_mst2slv_data[1];
end

//A kontroll regiszter TXCLR bitje (adsi FIFO trlse).
wire tx_fifo_clr = rst | (ctrl_reg_wr & s_mst2slv_data[2]);

//A kontroll regiszter RXCLR bitje (vteli FIFO trlse).
wire rx_fifo_clr = rst | (ctrl_reg_wr & s_mst2slv_data[3]);

//A kontroll regiszter vissaolvashat bitjei.
wire [7:0] ctrl_reg_dout = {6'b000000, rx_enable, tx_enable};


//******************************************************************************
//* A FIFO sttusz regiszter.                                                  *
//******************************************************************************
wire [7:0] stat_reg;
wire       tx_fifo_empty;
wire       tx_fifo_full;
wire       rx_fifo_empty;
wire       rx_fifo_full;

//A FIFO sttusz regiszter bitjei.
assign stat_reg[0]   = tx_fifo_empty;
assign stat_reg[1]   = ~tx_fifo_full;
assign stat_reg[2]   = ~rx_fifo_empty;
assign stat_reg[3]   = rx_fifo_full;
assign stat_reg[7:4] = 4'b0000;


//******************************************************************************
//* A megszakts engedlyez regiszter.                                       *
//******************************************************************************
reg  [3:0] ie_reg;
wire [7:0] ie_reg_dout = {4'b0000, ie_reg};

always @(posedge clk)
begin
   if (rst)
      ie_reg <= 4'b0000;
   else
      if (ie_reg_wr)
         ie_reg <= s_mst2slv_data[3:0];
end

//A megszaktskr kimenet meghajtsa.
assign irq = |(stat_reg[3:0] & ie_reg);


//******************************************************************************
//* Az olvassi adatbusz meghajtsa.                                           *
//******************************************************************************
wire [7:0] rx_data;
wire [3:0] dout_sel = {data_reg_rd, ie_reg_rd, stat_reg_rd, ctrl_reg_rd};

always @(*)
begin
   case (dout_sel)
      4'b0001: s_slv2mst_data <= ctrl_reg_dout;
      4'b0010: s_slv2mst_data <= stat_reg;
      4'b0100: s_slv2mst_data <= ie_reg_dout;
      4'b1000: s_slv2mst_data <= rx_data;
      default: s_slv2mst_data <= 8'd0;
   endcase
end


//******************************************************************************
//* ldetektls az USRT rajelen s az RXD vonal ksleltetse.                *
//******************************************************************************
(* shreg_extract = "no" *)
(* register_balancing = "no" *)
(* register_duplication = "no" *)
(* equivalent_register_removal = "no" *)
reg  [2:0] usrt_clk_samples;
wire       usrt_clk_rising  = (usrt_clk_samples[2:1] == 2'b01);
wire       usrt_clk_falling = (usrt_clk_samples[2:1] == 2'b10);

always @(posedge clk)
begin
   usrt_clk_samples <= {usrt_clk_samples[1:0], usrt_clk};
end

(* shreg_extract = "no" *)
(* register_balancing = "no" *)
(* register_duplication = "no" *)
(* equivalent_register_removal = "no" *)
reg  [1:0] rxd_samples;

always @(posedge clk)
begin
   rxd_samples <= {rxd_samples[0], usrt_rxd};
end


//******************************************************************************
//* USRT vev.                                                                 *
//******************************************************************************
//A vteli shiftregiszter.
reg  [9:0] rx_shr;
//A vteli shiftregisztert alapllapotba llt jel:
// - rendszer reset vagy
// - a vtel nincs engedlyezve vagy
// - a START bit 0
wire       rx_shr_rst = rst | ~rx_enable | ~rx_shr[0];
//A vett adat rvnyessgnek jelzse (START bit 0, STOP bit 1).
wire       rx_valid   = ~rx_shr[0] & rx_shr[9];

//A vteli shiftregiszter. Az USRT rajel lefut lre
//belptetjk az RXD vonal aktulis rtkt a regiszterbe.
always @(posedge clk)
begin
   if (rx_shr_rst)
      rx_shr <= 10'b11_1111_1111;
   else
      if (usrt_clk_falling)
         rx_shr <= {rxd_samples[1], rx_shr[9:1]};
end

//Vteli FIFO.
fifo rx_fifo(
   //rajel s reset.
   .clk(clk),                          //rajel
   .rst(rx_fifo_clr),                  //Reset jel
   
   //Adatvonalak.
   .data_in(rx_shr[8:1]),              //A FIFO-ba rand adat
   .data_out(rx_data),                 //A FIFO-bl kiolvasott adat
   
   //Vezrl bemenetek.
   .write(rx_valid),                   //rs engedlyez jel
   .read(data_reg_rd),                 //Olvass engedlyez jel
   
   //Sttusz kimenetek.
   .empty(rx_fifo_empty),              //A FIFO res
   .full(rx_fifo_full)                 //A FIFO tele van
);


//******************************************************************************
//* USRT ad.                                                                  *
//******************************************************************************
//Az adsi shiftregiszter.
reg  [8:0] tx_shr;
//Az elkldtt bitek szmllja.
reg  [3:0] tx_cnt;
//Az adsi FIFO-bl beolvasott adat.
wire [7:0] tx_fifo_dout;
//Az adsi shiftregiszter s a szmll betlt jele:
// - az ads engedlyezve van s
// - az elz keret utols bitjt is elkldtk (tx_cnt = 9) s
// - az adsi FIFO nem res
wire       tx_load = tx_enable & (tx_cnt == 4'd9) & ~tx_fifo_empty;

//Adsi FIFO.
fifo tx_fifo(
   //rajel s reset.
   .clk(clk),                          //rajel
   .rst(tx_fifo_clr),                  //Reset jel
   
   //Adatvonalak.
   .data_in(s_mst2slv_data),           //A FIFO-ba rand adat
   .data_out(tx_fifo_dout),            //A FIFO-bl kiolvasott adat
   
   //Vezrl bemenetek.
   .write(data_reg_wr),                //rs engedlyez jel
   .read(tx_load & usrt_clk_rising),   //Olvass engedlyez jel
   
   //Sttusz kimenetek.
   .empty(tx_fifo_empty),              //A FIFO res
   .full(tx_fifo_full)                 //A FIFO tele van
);

//Az adsi shiftregiszter. Az USRT rajel felfut lre
//trtnik a betlts vagy a lptets.
always @(posedge clk)
begin
   if (rst)
      tx_shr <= 9'b1_1111_1111;
   else
      if (usrt_clk_rising)
         if (tx_load)
            tx_shr <= {tx_fifo_dout, 1'b0};
         else
            tx_shr <= {1'b1, tx_shr[8:1]};
end

//Az elkldtt bitek szmllja.
always @(posedge clk)
begin
   if (rst)
      tx_cnt <= 4'd9;
   else
      if (usrt_clk_rising)
         if (tx_load)
            tx_cnt <= 4'd0;
         else
            if (tx_cnt != 4'd9)
               tx_cnt <= tx_cnt + 4'd1;
end

//Az USRT interfsz TXD vonalnak meghajtsa.
assign usrt_txd = tx_shr[0];


endmodule
