`timescale 1ns / 1ps

//******************************************************************************
//* 1024 x 768 @ 60 Hz VGA megjelent modul grafikus s karakteres zemmd    *
//* tmogatssal.                                                              *
//*                                                                            *
//* A programozi fellet:                                                     *
//*                                                                            *
//* Cm         Tpus   Bitek                                                  *
//* BASEADDR+0          VGA kontroll/sttusz regiszter                         *
//*             WR      -     -     -     -     -     INC   MODE  EN           *
//*             RD      VBL   IRQ   0     0     0     INC   MODE  EN           *
//* BASEADDR+1          Megszakts engedlyez regiszter                      *
//*             WR      -     -     -     -     -     -     -     VBIE         *
//*             RD      0     0     0     0     0     0     0     VBIE         *
//* BASEADDR+2          Megszakts flag regiszter                             *
//*             W1C     -     -     -     -     -     -     -     VBIF         *
//*             RD      0     0     0     0     0     0     0     VBIF         *
//* BASEADDR+3  RD/WR   Adatregiszter                                          *
//* BASEADDR+4  RD/WR   X-koordinta regiszter                                 *
//* BASEADDR+5  RD/WR   Y-koordinta regiszter                                 *
//*                                                                            *
//* A grafikus zemmd (MODE=0) felbontsa 256x192 pixel, a karakteres zemmd *
//* (MODE=1) felbontsa pedig 32x24 karakter. A koordinta regiszterekbe ennek *
//* megfelel rtkek randk. A (0, 0) a bal fels sarok koordintja.        *
//*                                                                            *
//* Grafikus zemmd esetn egy adatregiszter rssal egy pixel vezrelhet.   *
//* A pixel sznt a bert bjt als hrom bitje hatrozza meg.                *
//*                                                                            *
//* Karakteres md esetn egy karakterhez kt adatregiszter elrs tartozik:   *
//*                  D7    | D6 | D5  D4  D3 | D2   D1   D0                    *
//* 1. rs/olvass: a megjelentend karakter ASCII kdja                     *
//* 2. rs/olvass: BLINK | -  | httrszn | karakterszn                    *
//*                                                                            *
//* Az adatregiszter elrs sorn az INC bit rtknek megfelelen vltoznak a *
//* koordintk:                                                               *
//* INC=0: vzszintes rs/olvass, az X-koordinta nvekszik elszr          *
//* INC=1: fggleges rs/olvass, az Y-koordinta nvekszik elszr          *
//******************************************************************************
module vga_display #(
   //A perifria bziscme.
   parameter BASEADDR = 8'hff
) (
   //rajel s reset.
   input  wire       clk,              //rajel
   input  wire       rst,              //Reset jel
   
   //A slave busz interfsz jelei.
   input  wire [7:0] s_mst2slv_addr,   //Cmbusz
   input  wire       s_mst2slv_wr,     //rs engedlyez jel
   input  wire       s_mst2slv_rd,     //Olvass engedlyez jel
   input  wire [7:0] s_mst2slv_data,   //rsi adatbusz
   output reg  [7:0] s_slv2mst_data,   //Olvassi adatbusz
   
   //Megszaktskr kimenet.
   output wire       irq,
   
   //A VGA interfsz jelei.
   output wire       vga_enabled,      //A VGA interfsz engedlyezett
   output reg  [5:0] rgb_out,          //Szn adatok
   output wire       hsync_out,        //Horizontlis szinkronjel
   output wire       vsync_out         //Vertiklis szinkronjel
);

//******************************************************************************
//* Cmdekdols.                                                              *
//******************************************************************************
//A perifria kivlaszt jele.
wire psel = ((s_mst2slv_addr >> 3) == (BASEADDR >> 3));

//A kontroll/sttusz regiszter rsnak s olvassnak jelzse.
wire ctrl_reg_wr = psel & s_mst2slv_wr & (s_mst2slv_addr[2:0] == 3'b000);
wire stat_reg_rd = psel & s_mst2slv_rd & (s_mst2slv_addr[2:0] == 3'b000);

//A megszakts engedlyez regiszter rsnak s olvassnak jelzse.
wire ie_reg_wr = psel & s_mst2slv_wr & (s_mst2slv_addr[2:0] == 3'b001);
wire ie_reg_rd = psel & s_mst2slv_rd & (s_mst2slv_addr[2:0] == 3'b001);

//A megszakts flag regiszter rsnak s olvassnak jelzse.
wire if_reg_wr = psel & s_mst2slv_wr & (s_mst2slv_addr[2:0] == 3'b010);
wire if_reg_rd = psel & s_mst2slv_rd & (s_mst2slv_addr[2:0] == 3'b010);

//Az adatregiszter rsnak s olvassnak jelzse.
wire data_reg_wr = psel & s_mst2slv_wr & (s_mst2slv_addr[2:0] == 3'b011);
wire data_reg_rd = psel & s_mst2slv_rd & (s_mst2slv_addr[2:0] == 3'b011);

//A cmszmll rsnak s olvassnak jelzse.
wire xcnt_reg_wr = psel & s_mst2slv_wr & (s_mst2slv_addr[2:0] == 3'b100);
wire xcnt_reg_rd = psel & s_mst2slv_rd & (s_mst2slv_addr[2:0] == 3'b100);
wire ycnt_reg_wr = psel & s_mst2slv_wr & (s_mst2slv_addr[2:0] == 3'b101);
wire ycnt_reg_rd = psel & s_mst2slv_rd & (s_mst2slv_addr[2:0] == 3'b101);


//******************************************************************************
//* Kontroll/sttusz regiszter.                                                *
//******************************************************************************
reg [2:0] ctrl_reg;

always @(posedge clk)
begin
   if (rst)
      ctrl_reg <= 3'b000;
   else
      if (ctrl_reg_wr)
         ctrl_reg <= s_mst2slv_data[2:0];
end

//A kontroll regiszter bitjei.
assign vga_enabled = ctrl_reg[0];
wire   disp_mode   = ctrl_reg[1];
wire   adr_inc_sel = ctrl_reg[2];

//A sttusz regiszter bitjei.
wire [7:0] stat_reg;
wire       v_blank;

assign stat_reg[2:0] = ctrl_reg;
assign stat_reg[5:3] = 3'b000;
assign stat_reg[6]   = irq;
assign stat_reg[7]   = v_blank;


//******************************************************************************
//* Megszakts engedlyez regiszter.                                         *
//******************************************************************************
reg ie_reg;

always @(posedge clk)
begin
   if (rst)
      ie_reg <= 1'b0;
   else
      if (ie_reg_wr)
         ie_reg <= s_mst2slv_data[0];
end


//******************************************************************************
//* Megszakts flag regiszter.                                                *
//******************************************************************************
reg  if_reg;
wire v_blank_begin;
wire v_blank_end;

always @(posedge clk)
begin
   if (rst)
      if_reg <= 1'b0;
   else
      if (v_blank_begin)
         if_reg <= 1'b1;
      else
         if (v_blank_end || (if_reg_wr && s_mst2slv_data[0]))
            if_reg <= 1'b0;
end

//A megszaktskr kimenet meghajtsa.
assign irq = ie_reg & if_reg;


//******************************************************************************
//* Adatregiszter (video memria).                                             *
//******************************************************************************
reg  [15:0] vram_cpu_addr;
wire [7:0]  vram_cpu_dout;

reg  [15:0] vram_vga_addr;
wire [2:0]  vram_rgb_data;
wire [15:0] vram_chr_data;

video_memory video_memory(
   //rajel.
   .clk(clk),
   
   //zemmd kivlaszt jel (0: grafikus, 1: karakteres).
   .mode(disp_mode),
   
   //rsi/olvassi port a CPU fel.
   .cpu_addr(vram_cpu_addr),           //Cmbusz
   .cpu_write(data_reg_wr),            //rs engedlyez jel
   .cpu_din(s_mst2slv_data),           //rsi adatbusz
   .cpu_dout(vram_cpu_dout),           //Olvassi adatbusz
   
   //Olvassi port a VGA modul fel.
   .vga_addr(vram_vga_addr),           //Olvassi cm
   .vga_rgb_data(vram_rgb_data),       //Adat (grafikus md)
   .vga_char_data(vram_chr_data)       //Adat (karakteres md)
);


//******************************************************************************
//* Cmszmll.                                                               *
//******************************************************************************
//Szmll az X-koordinthoz.
reg  [7:0] x_cnt;
reg        x_cnt_en;
wire       x_cnt_max = (x_cnt == ((disp_mode) ? 8'd31 : 8'd255));

always @(posedge clk)
begin
   if (rst)
      x_cnt <= 8'd0;
   else
      if (xcnt_reg_wr)
         x_cnt <= s_mst2slv_data;
      else
         if (x_cnt_en)
            if (x_cnt_max)
               x_cnt <= 8'd0;
            else
               x_cnt <= x_cnt + 8'd1;
end

//Szmll az X-koordinthoz.
reg  [7:0] y_cnt;
reg        y_cnt_en;
wire       y_cnt_max = (y_cnt == ((disp_mode) ? 8'd23 : 8'd191));

always @(posedge clk)
begin
   if (rst)
      y_cnt <= 8'd0;
   else
      if (ycnt_reg_wr)
         y_cnt <= s_mst2slv_data;
      else
         if (y_cnt_en)
            if (y_cnt_max)
               y_cnt <= 8'd0;
            else
               y_cnt <= y_cnt + 8'd1;
end

//Karakteres zemmd esetn a cm LSb-je.
reg  chr_addr_lsb;
wire data_reg_rdwr = data_reg_wr | data_reg_rd;

always @(posedge clk)
begin
   if (rst || xcnt_reg_wr)
      chr_addr_lsb <= 1'b0;
   else
      if (data_reg_rdwr)
         chr_addr_lsb <= ~chr_addr_lsb;
end

//Az X szmll engedlyez jelnek ellltsa.
always @(*)
begin
   if (disp_mode)
      if (adr_inc_sel)
         //Karakteres zemmd, fggleges rs/olvass.
         x_cnt_en <= data_reg_rdwr & chr_addr_lsb & y_cnt_max;
      else
         //Karakteres zemmd, vzszintes rs/olvass.
         x_cnt_en <= data_reg_rdwr & chr_addr_lsb;
   else
      if (adr_inc_sel)
         //Grafikus zemmd, fggleges rs/olvass.
         x_cnt_en <= data_reg_rdwr & y_cnt_max;
      else
         //Grafikus zemmd, vzszintes rs/olvass.
         x_cnt_en <= data_reg_rdwr;
end

//Az Y szmll engedlyez jelnek ellltsa.
always @(*)
begin
   if (disp_mode)
      if (adr_inc_sel)
         //Karakteres zemmd, fggleges rs/olvass.
         y_cnt_en <= data_reg_rdwr & chr_addr_lsb;
      else
         //Karakteres zemmd, vzszintes rs/olvass.
         y_cnt_en <= data_reg_rdwr & chr_addr_lsb & x_cnt_max;
   else
      if (adr_inc_sel)
         //Grafikus zemmd, fggleges rs/olvass.
         y_cnt_en <= data_reg_rdwr;
      else
         //Grafikus zemmd, vzszintes rs/olvass.
         y_cnt_en <= data_reg_rdwr & x_cnt_max;
end

//A CPU oldali video memria cm ellltsa.
always @(*)
begin
   if (disp_mode)
      vram_cpu_addr <= {5'd0, y_cnt[4:0], x_cnt[4:0], chr_addr_lsb};
   else
      vram_cpu_addr <= {y_cnt, x_cnt};
end


//******************************************************************************
//* A processzor olvassi adatbusznak meghajtsa. Az olvassi adatbuszra csak *
//* az olvass ideje alatt kapcsoljuk r a krt rtket, egybknt egy inaktv *
//* nulla rtk jelenik meg rajta (elosztott busz multiplexer funkci).        *
//******************************************************************************
wire [5:0] dout_sel;

assign dout_sel[0] = stat_reg_rd;
assign dout_sel[1] = ie_reg_rd;
assign dout_sel[2] = if_reg_rd;
assign dout_sel[3] = data_reg_rd;
assign dout_sel[4] = xcnt_reg_rd;
assign dout_sel[5] = ycnt_reg_rd;

always @(*)
begin
   case (dout_sel)
      6'b000001: s_slv2mst_data <= stat_reg;
      6'b000010: s_slv2mst_data <= {7'd0, ie_reg};
      6'b000100: s_slv2mst_data <= {7'd0, if_reg};
      6'b001000: s_slv2mst_data <= vram_cpu_dout;
      6'b010000: s_slv2mst_data <= x_cnt;
      6'b100000: s_slv2mst_data <= y_cnt;
      default  : s_slv2mst_data <= 8'd0;
   endcase
end


//******************************************************************************
//* VGA idzts genertor.                                                    *
//******************************************************************************
wire [8:0] h_cnt;
wire [9:0] v_cnt;
wire       h_blank;
wire       h_sync;
wire       v_sync;

vga_timing vga_timing(
   //rajel s reset.
   .clk(clk),                          //rajel
   .rst(rst),                          //Reset jel
   
   //Horizontlis s vertiklis szmllk.
   .h_cnt(h_cnt),                      //Horizontlis szmll
   .v_cnt(v_cnt),                      //Vertiklis szmll
   
   //Szinkron- s kioltjelek. 
   .h_blank(h_blank),                  //Horizontlis kioltjel
   .v_blank(v_blank),                  //Vertiklis kioltjel
   .v_blank_begin(v_blank_begin),      //A vertiklis visszafuts kezdete
   .v_blank_end(v_blank_end),          //A vertiklis visszafuts vge
   .h_sync(h_sync),                    //Horizontlis szinkronjel
   .v_sync(v_sync)                     //Vertiklis szinkronjel
);

//A kiolt jel ellltsa.
wire blank = h_blank | v_blank;

//A videomemria VGA oldali cmnek ellltsa.
always @(*)
begin
   if (disp_mode)
      //Karakteres zemmd.
      vram_vga_addr <= {5'd0, v_cnt[9:5], h_cnt[7:3], 1'd0};
   else
      //Grafikus zemmd.
      vram_vga_addr <= {v_cnt[9:2], h_cnt[7:0]};
end


//******************************************************************************
//* Karakter ROM.                                                              *
//******************************************************************************
(* rom_style = "block" *)
reg chr_rom [16383:0];
reg chr_rom_dout;

initial
begin
   $readmemh("src/Peripherals/VGA/font_data.txt", chr_rom, 0, 16383);
end

//A karakter ROM cm als 6 bitjnek ksleltetse. Ez azrt szksges, mert
//a video memria blokk-RAM-mal van megvalstva, gy olvass esetn 1 rajel
//temnyit ksleltet.
reg  [5:0]  chr_rom_adrl;
wire [13:0] chr_rom_addr = {vram_chr_data[7:0], chr_rom_adrl}; 

always @(posedge clk)
begin
   chr_rom_adrl = {v_cnt[4:2], h_cnt[2:0]};
end

//A karakrer ROM kimenete.
always @(posedge clk)
begin
   chr_rom_dout <= chr_rom[chr_rom_addr];
end


//******************************************************************************
//* A VGA interfsz kimeneteinek meghajtsa.                                   *
//******************************************************************************
//A szksges jelek ksleltetse a blokk-RAM-ok szmnak megfelelen.
reg [7:0] chr_reg;
reg [2:0] rgb_reg;
reg [1:0] blank_reg;
reg [2:0] h_sync_reg;
reg [2:0] v_sync_reg;

always @(posedge clk)
begin
   chr_reg    <= vram_chr_data[15:8];
   rgb_reg    <= vram_rgb_data;
   blank_reg  <= {blank_reg[0], blank};
   h_sync_reg <= {h_sync_reg[1:0], h_sync};
   v_sync_reg <= {v_sync_reg[1:0], v_sync};
end

//rajel oszt a karakterek villogtatshoz.
reg [5:0] blink_cnt;

always @(posedge clk)
begin
   if (rst)
      blink_cnt <= 6'd0;
   else
      if (v_blank_begin)
         blink_cnt <= blink_cnt + 6'd1;
end

//Az RGB kimenet meghajtsa.
always @(posedge clk)
begin
   if (blank_reg[1])
      rgb_out <= 6'b00_00_00;
   else
      if (disp_mode)
         if (chr_rom_dout && (~chr_reg[7] || ~blink_cnt[5]))
            rgb_out <= {chr_reg[2], chr_reg[2], chr_reg[1], chr_reg[1], chr_reg[0], chr_reg[0]};
         else
            rgb_out <= {chr_reg[5], chr_reg[5], chr_reg[4], chr_reg[4], chr_reg[3], chr_reg[3]};
      else
         rgb_out <= {rgb_reg[2], rgb_reg[2], rgb_reg[1], rgb_reg[1], rgb_reg[0], rgb_reg[0]};
end

//A szinkronjelek meghajtsa.
assign hsync_out = h_sync_reg[2];
assign vsync_out = v_sync_reg[2];

endmodule
